@@ -12,7 +12,7 @@ use base 'PDF::API2::Basic::PDF::Objind';
 
 use strict;
 
-our $VERSION = '2.039'; # VERSION
+our $VERSION = '2.040'; # VERSION
 
 =head1 NAME
 
@@ -191,12 +191,12 @@ sub as_pdf {
         return "<FEFF$str>";
     }
     else {
-        if ($str =~ m/[^\n\r\t\b\f\040-\176\200-\377]/oi) {
-            $str =~ s/(.)/sprintf('%02X', ord($1))/oge;
+        if ($str =~ m/[^\n\r\t\b\f\040-\176\200-\377]/) {
+            $str =~ s/(.)/sprintf('%02X', ord($1))/sge;
             return "<$str>";
         }
         else {
-            $str =~ s/([\n\r\t\b\f\\()])/\\$out_trans{$1}/ogi;
+            $str =~ s/([\n\r\t\b\f\\()])/\\$out_trans{$1}/g;
             return "($str)";
         }
     }
