@@ -5,7 +5,7 @@ use base 'PDF::API2::Resource';
 use strict;
 use warnings;
 
-our $VERSION = '2.035'; # VERSION
+our $VERSION = '2.040'; # VERSION
 
 use Compress::Zlib;
 use Encode qw(:all);
@@ -110,7 +110,12 @@ sub tounicodemap {
                 $stream .= qq|endbfrange\n|;
                 $stream .= qq|$i beginbfrange\n|;
             }
-            $stream .= sprintf(qq|<%04x> <%04x> <%04x>\n|, $j, $j, $self->uniByCId($j));
+
+            # Default to 0000 if uniByCId returns undef in order to match
+            # previous behavior minus an uninitialized value warning.  It's
+            # worth looking into what should be happening here, since this may
+            # not be the correct behavior.
+            $stream .= sprintf(qq|<%04x> <%04x> <%04x>\n|, $j, $j, ($self->uniByCId($j) // 0));
         }
         $stream .= "endbfrange\n";
     }
@@ -119,7 +124,7 @@ sub tounicodemap {
         $stream .= qq|1 begincodespacerange\n<00> <FF>\nendcodespacerange\n|;
         $stream .= qq|256 beginbfchar\n|;
         for (my $j = 0; $j < 256; $j++) {
-            $stream .= sprintf(qq|<%02X> <%04X>\n|, $j, $self->uniByEnc($j));
+            $stream .= sprintf(qq|<%02X> <%04X>\n|, $j, $self->uniByEnc($j) // 0);
         }
         $stream .= qq|endbfchar\n|;
     }
